@@ -23,6 +23,7 @@ import collections
 import hashlib
 import html.entities
 import logging
+import os
 import shelve
 import ssl
 import tempfile
@@ -32,13 +33,18 @@ from http.client import HTTPSConnection
 from urllib.parse import quote_plus
 from xml.dom import Node, minidom
 
-import pkg_resources
+try:
+    # Python 3.8+
+    import importlib.metadata as importlib_metadata
+except ImportError:
+    # Python 3.7 and lower
+    import importlib_metadata
 
 __author__ = "Amr Hassan, hugovk, Mice Pápai"
 __copyright__ = "Copyright (C) 2008-2010 Amr Hassan, 2013-2021 hugovk, 2017 Mice Pápai"
 __license__ = "apache2"
 __email__ = "amr.hassan@gmail.com"
-__version__ = pkg_resources.get_distribution(__name__).version
+__version__ = importlib_metadata.version(__name__)
 
 
 # 1 : This error does not exist
@@ -421,7 +427,8 @@ class _Network:
         """
 
         if not file_path:
-            file_path = tempfile.mktemp(prefix="pylast_tmp_")
+            self.cache_backend = _ShelfCacheBackend.create_shelf()
+            return
 
         self.cache_backend = _ShelfCacheBackend(file_path)
 
@@ -782,8 +789,11 @@ class LibreFMNetwork(_Network):
 class _ShelfCacheBackend:
     """Used as a backend for caching cacheable requests."""
 
-    def __init__(self, file_path=None):
-        self.shelf = shelve.open(file_path)
+    def __init__(self, file_path=None, flag=None):
+        if flag is not None:
+            self.shelf = shelve.open(file_path, flag=flag)
+        else:
+            self.shelf = shelve.open(file_path)
         self.cache_keys = set(self.shelf.keys())
 
     def __contains__(self, key):
@@ -799,6 +809,12 @@ class _ShelfCacheBackend:
         self.cache_keys.add(key)
         self.shelf[key] = xml_string
 
+    @classmethod
+    def create_shelf(cls):
+        file_descriptor, file_path = tempfile.mkstemp(prefix="pylast_tmp_")
+        os.close(file_descriptor)
+        return cls(file_path=file_path, flag="n")
+
 
 class _Request:
     """Representing an abstract web service operation."""
@@ -888,6 +904,9 @@ class _Request:
         if self.network.limit_rate:
             self.network._delay_call()
 
+        username = self.params.pop("username", None)
+        username = f"?username={username}" if username is not None else ""
+
         data = []
         for name in self.params.keys():
             data.append("=".join((name, quote_plus(_string(self.params[name])))))
@@ -911,20 +930,25 @@ class _Request:
             try:
                 conn.request(
                     method="POST",
-                    url="https://" + host_name + host_subdir,
+                    url=f"https://{host_name}{host_subdir}{username}",
                     body=data,
                     headers=headers,
                 )
             except Exception as e:
-                raise NetworkError(self.network, e)
+                raise NetworkError(self.network, e) from e
 
         else:
             conn = HTTPSConnection(context=SSL_CONTEXT, host=host_name)
 
             try:
-                conn.request(method="POST", url=host_subdir, body=data, headers=headers)
+                conn.request(
+                    method="POST",
+                    url=f"{host_subdir}{username}",
+                    body=data,
+                    headers=headers,
+                )
             except Exception as e:
-                raise NetworkError(self.network, e)
+                raise NetworkError(self.network, e) from e
 
         try:
             response = conn.getresponse()
@@ -937,7 +961,7 @@ class _Request:
                 )
             response_text = _unicode(response.read())
         except Exception as e:
-            raise MalformedResponseError(self.network, e)
+            raise MalformedResponseError(self.network, e) from e
 
         try:
             self._check_response_for_errors(response_text)
@@ -961,7 +985,7 @@ class _Request:
         try:
             doc = minidom.parseString(_string(response).replace("opensearch:", ""))
         except Exception as e:
-            raise MalformedResponseError(self.network, e)
+            raise MalformedResponseError(self.network, e) from e
 
         e = doc.getElementsByTagName("lfm")[0]
         # logger.debug(doc.toprettyxml())
@@ -1042,9 +1066,6 @@ class SessionKeyGenerator:
 
         if url in self.web_auth_tokens.keys():
             token = self.web_auth_tokens[url]
-        else:
-            # This will raise a WSError if token is blank or unauthorized
-            token = token
 
         request = _Request(self.network, "auth.getSession", {"token": token})
 
@@ -1339,7 +1360,7 @@ class _Taggable(_BaseObject):
             self.remove_tag(tag)
 
     def clear_tags(self):
-        """Clears all the user-set tags. """
+        """Clears all the user-set tags."""
 
         self.remove_tags(*(self.get_tags()))
 
@@ -1397,7 +1418,13 @@ class _Taggable(_BaseObject):
         return seq
 
 
-class WSError(Exception):
+class PyLastError(Exception):
+    """Generic exception raised by PyLast"""
+
+    pass
+
+
+class WSError(PyLastError):
     """Exception related to the Network web service"""
 
     def __init__(self, network, status, details):
@@ -1441,7 +1468,7 @@ class WSError(Exception):
         return self.status
 
 
-class MalformedResponseError(Exception):
+class MalformedResponseError(PyLastError):
     """Exception conveying a malformed response from the music network."""
 
     def __init__(self, network, underlying_error):
@@ -1454,7 +1481,7 @@ class MalformedResponseError(Exception):
         )
 
 
-class NetworkError(Exception):
+class NetworkError(PyLastError):
     """Exception conveying a problem in sending a request to Last.fm"""
 
     def __init__(self, network, underlying_error):
@@ -1494,7 +1521,9 @@ class _Opus(_Taggable):
             self.artist = Artist(artist, self.network)
 
         self.title = title
-        self.username = username
+        self.username = (
+            username if username else network.username
+        )  # Default to current user
         self.info = info
 
     def __repr__(self):
@@ -1678,7 +1707,7 @@ class Artist(_Taggable):
         self.info = info
 
     def __repr__(self):
-        return "pylast.Artist({}, {})".format(repr(self.get_name()), repr(self.network))
+        return f"pylast.Artist({repr(self.get_name())}, {repr(self.network)})"
 
     def __unicode__(self):
         return str(self.get_name())
@@ -1774,9 +1803,14 @@ class Artist(_Taggable):
         else:
             params = None
 
-        return self._extract_cdata_from_request(
-            self.ws_prefix + ".getInfo", section, params
-        )
+        try:
+            bio = self._extract_cdata_from_request(
+                self.ws_prefix + ".getInfo", section, params
+            )
+        except IndexError:
+            bio = None
+
+        return bio
 
     def get_bio_published_date(self):
         """Returns the date on which the artist's biography was published."""
@@ -1862,7 +1896,7 @@ class Country(_BaseObject):
         self.name = name
 
     def __repr__(self):
-        return "pylast.Country({}, {})".format(repr(self.name), repr(self.network))
+        return f"pylast.Country({repr(self.name)}, {repr(self.network)})"
 
     @_string_output
     def __str__(self):
@@ -1878,7 +1912,7 @@ class Country(_BaseObject):
         return {"country": self.get_name()}
 
     def get_name(self):
-        """Returns the country name. """
+        """Returns the country name."""
 
         return self.name
 
@@ -1940,7 +1974,7 @@ class Library(_BaseObject):
             self.user = User(user, self.network)
 
     def __repr__(self):
-        return "pylast.Library({}, {})".format(repr(self.user), repr(self.network))
+        return f"pylast.Library({repr(self.user)}, {repr(self.network)})"
 
     @_string_output
     def __str__(self):
@@ -1986,7 +2020,7 @@ class Tag(_Chartable):
         self.name = name
 
     def __repr__(self):
-        return "pylast.Tag({}, {})".format(repr(self.name), repr(self.network))
+        return f"pylast.Tag({repr(self.name)}, {repr(self.network)})"
 
     @_string_output
     def __str__(self):
@@ -2002,7 +2036,7 @@ class Tag(_Chartable):
         return {self.ws_prefix: self.get_name()}
 
     def get_name(self, properly_capitalized=False):
-        """Returns the name of the tag. """
+        """Returns the name of the tag."""
 
         if properly_capitalized:
             self.name = _extract(
@@ -2125,12 +2159,12 @@ class Track(_Opus):
         return Album(_extract(node, "artist"), _extract(node, "title"), self.network)
 
     def love(self):
-        """Adds the track to the user's loved tracks. """
+        """Adds the track to the user's loved tracks."""
 
         self._request(self.ws_prefix + ".love")
 
     def unlove(self):
-        """Remove the track to the user's loved tracks. """
+        """Remove the track to the user's loved tracks."""
 
         self._request(self.ws_prefix + ".unlove")
 
@@ -2196,7 +2230,7 @@ class User(_Chartable):
         self.name = user_name
 
     def __repr__(self):
-        return "pylast.User({}, {})".format(repr(self.name), repr(self.network))
+        return f"pylast.User({repr(self.name)}, {repr(self.network)})"
 
     @_string_output
     def __str__(self):
@@ -2235,7 +2269,7 @@ class User(_Chartable):
         return self.name
 
     def get_friends(self, limit=50, cacheable=False, stream=False):
-        """Returns a list of the user's friends. """
+        """Returns a list of the user's friends."""
 
         def _get_friends():
             for node in _collect_nodes(
@@ -2517,8 +2551,7 @@ class User(_Chartable):
 
         params = self._get_params()
         params["period"] = period
-        if limit:
-            params["limit"] = limit
+        params["limit"] = limit
 
         return self._get_things("getTopTracks", Track, params, cacheable, stream=stream)
 
@@ -2580,7 +2613,7 @@ class User(_Chartable):
         return self.network._get_url(domain_name, "user") % {"name": name}
 
     def get_library(self):
-        """Returns the associated Library object. """
+        """Returns the associated Library object."""
 
         return Library(self, self.network)
 
@@ -2778,7 +2811,7 @@ def _collect_nodes(limit, sender, method_name, cacheable, params=None, stream=Fa
                     break  # success
                 except Exception as e:
                     if tries >= 3:
-                        raise e
+                        raise PyLastError() from e
                     # Wait and try again
                     time.sleep(1)
                     tries += 1
@@ -2795,7 +2828,7 @@ def _collect_nodes(limit, sender, method_name, cacheable, params=None, stream=Fa
                     main.getAttribute("totalPages") or main.getAttribute("totalpages")
                 )
             else:
-                raise Exception("No total pages attribute")
+                raise PyLastError("No total pages attribute")
 
             for node in main.childNodes:
                 if not node.nodeType == xml.dom.Node.TEXT_NODE and (
@@ -2910,9 +2943,6 @@ def _number(string):
 
 
 def _unescape_htmlentity(string):
-
-    # string = _unicode(string)
-
     mapping = html.entities.name2codepoint
     for key in mapping:
         string = string.replace("&%s;" % key, chr(mapping[key]))
